__all__ = ["spectrometer_sensitivity"]


# standard library
from typing import List, Union, Dict


# dependent packages
import numpy as np
import pandas as pd
from .atmosphere import eta_atm_func
from .instruments import eta_Al_ohmic_850, photon_NEP2_kid, window_trans, average_over_filterbank, eta_ruze
from .physics import johnson_nyquist_psd, rad_trans, T_from_psd, c, h, k
from .filterbank import generateFilterbankFromR

from .cascade import get_cascade

# type aliases
ArrayLike = Union[np.ndarray, List[float], List[int], float, int]

# main functions
def calculator(
    cascade_list: List[Dict[any, any]],
    telescope_dict: Dict[any, any],
    instrument_dict: Dict[any, any],
    pwv: float = 0.5,
    Tb_cmb: ArrayLike = 2.725,
    Tp_atm: float = 273,
    snr: float = 1.0,
    obs_hours: float = 2*8.0,
    on_source_fraction: float = 0.4 * 0.9,
    on_off: bool = True
):
    """Calculate the sensitivity of a spectrometer.

    Parameters which are functions of frequency can be a vector (see Parameters).
    Output is a pandas DataFrame which containts results of simulation (see Returns).

    Parameters
    ----------
    cascade_list
        List containing cascade stage dictionaries.
    telescope_dict
        Dictionary containing telescope specifications.
    instrument_dict
        Dictionary containing instrument specifications.
    pwv
        Precipitable water vapour. Units: mm.
    Tb_cmb
        Brightness temperature of CMB. Units: K.
    Tp_atm
        Physical temperature of atmosphere. Units: K.
    snr
        Target signal to noise to be reached (for calculating the MDLF). Units: None.
    obs_hours
        Observing hours, including off-source time and the slew overhead
        between on- and off-source. Units: hours.
    on_source_fraction
        Fraction of the time on source (between 0. and 1.). Units: None.
    on_off
        If the observation involves on_off chopping, then the SNR degrades
        by sqrt(2) because the signal difference includes the noise twice.

    Returns
    ----------
    Dictionary with output. It contains the following fields:
        - "F_KID"         : Filter frequencies in Hz.
        - "F_sky"         : Sky frequencies in Hz.
        - "EL"            : Telescope elevation for calculation.
        - "PWV"           : Precipitable water vapor for calculation.
        - "R"             : Resolving power of filterbank.
        - "filterbank"    : 2D array containing filterbank.
        - "W_F_spec"      : Equivalent width of each filter, coupling to a spectral line.
        - "W_F_cont"      : Equivalent width of each filter, coupling to continuum (so including out-of-band loading).
        - "eta_atm"       : Transmission of atmopsphere, averaged over filterbank response.
        - "eta_ap"        : Aperture efficiency, which is taper x Ruze x atmospherically coupled efficiencies, averaged over filterbank.
        - "eta_fwd"       : Forward efficiency, i.e. fraction of beam coupling to sky, averaged over filterbank response.
        - "eta_sw"        : Total coupling of source to cryostat window, averaged over filterbank response.
        - "eta_window"    : Total transmission efficiency of cryostat window, averaged over filterbank response.
        - "eta_inst"      : Total instrument efficiency, averaged over filterbank response.
        - "Tb_sky"        : Brightness temperature of the sky, averaged over filterbank response.
        - "psd_KID"       : Power spectral density entering each KID, averaged over filterbank response.
        - "P_KID"         : Power entering each KID.
        - "NEP_KID"       : Noise equivalent power for each KID, at the KID itself.
        - "NEP_inst"      : Noise equivalent power for each KID, evaluated at the cryostat window.
        - "NET_sky"       : Noise equivalent temperature, evaluated before atmosphere. Units: K s^0.5
        - "NEFD_line"     : Noise equivalent flux density for each KID, coupling to a spectral line.
        - "NEFD_continuum": Noise equivalent flux density for each KID, coupling to continuum (so including out-of-band loading).
        - "MDLF"          : Minimum detectable line flux, for each KID.
        - "equivalent_Trx": Equivalent receiver noise temperature, for each KID, at the cryostat window.
        - "n_ph"          : Photon occupation number, defined as number of photons arriving per coherence time, for each KID.
        - "W_F_cont"      : Continuum equivalent width.
    """

    # Unpacking telescope dictionary
    D_tel = telescope_dict["D_tel"]
    s_rms = telescope_dict["s_rms"]
    eta_taper = telescope_dict["eta_taper"]
    EL = telescope_dict["EL0"]

    # Unpacking instrument dictionary
    F = instrument_dict["f_ch"]
    R = instrument_dict["R"]
    eta_IBF = instrument_dict["eta_IBF"]
    KID_excess_noise_factor = 1 if instrument_dict.get("KID_excess_noise_factor") is None else instrument_dict.get("KID_excess_noise_factor")
    
    # Equivalent Bandwidth of 1 channel.
    # Used for calculating loading and coupling to a continuum source
    W_F_cont = F / R / eta_IBF
    # Used for calculating coupling to a line source,
    # with a linewidth not wider than the filter channel
    W_F_spec = F / R

    # #############################################################
    # 1. Calculating loading power absorbed by the KID, and the NEP
    # #############################################################
    if not hasattr(F, "__len__"):
        F = np.array([F])
    
    if not hasattr(pwv, "__len__"):
        pwv = np.array([pwv])

    if R is not None:
        filterbank, F_sky, dF_sky = generateFilterbankFromR(R, F)

    else: # Read from file
        pass

    eta_atm = np.squeeze(eta_atm_func(F=F_sky, pwv=pwv, EL=EL))

    eta_cascade, psd_cascade, use_for_eta_inst, use_for_eta_ap = get_cascade(cascade_list, F_sky)
                
    psd_in = johnson_nyquist_psd(F_sky, Tb_cmb)
    psd_atm = johnson_nyquist_psd(F_sky, Tp_atm)
    
    # Start on branch 1, increment each time eta couples to "atmosphere". Set to zero when cryo window is encountered.
    branch_fwd = 1
    index_branches = []

    # Initialise array for eta_inst and eta_ap
    eta_inst = np.ones(F_sky.size)
    eta_ap = np.ones(F_sky.size)
    
    # Do first stage outside loop, is always the same anyways...
    psd_running = eta_atm * psd_in + (1 - eta_atm) *  psd_atm
    psd_sky = average_over_filterbank(psd_running, filterbank)

    psd_in_front_of_cryo_set = False

    # Helper list for storing intermediate transmitted psd
    psd_l = []

    psd_l.append(average_over_filterbank(psd_running, filterbank))

    eta_window_set = False

    for idx, (eta_stage, psd_stage) in enumerate(zip(eta_cascade, psd_cascade)):
        if isinstance(psd_stage, str):
            psd_stage = (1 - eta_atm) * psd_atm
            branch_fwd += 1 
        
        if use_for_eta_ap[idx]:
            eta_ap *= eta_stage     

        if use_for_eta_inst[idx]:
            # If stage is inside cryostat (including window), it counts towards eta_inst
            eta_inst *= eta_stage 
        
            # Also check, if this is first stage inside instrument -> assign to eta_window
            if eta_window_set == False:
                eta_window = eta_stage
                eta_window_set = True

            # Also, stop incorporating eta into eta_fwd
            branch_fwd = 0

            if psd_in_front_of_cryo_set == False:
                # Save psd up till now for Trx calculation
                psd_in_front_of_cryo = psd_running
                psd_in_front_of_cryo_set = True

        if idx == len(eta_cascade) - 1:
            eta_stage = np.tile(eta_stage, F.size).reshape(filterbank.shape) * filterbank

        psd_running = rad_trans(psd_running, psd_stage, eta_stage)

        index_branches.append(branch_fwd)

    # Absorb eta_window in eta_inst as well
    eta_inst *= eta_window

    psd_KID = np.nansum(psd_running, axis=0) / np.nansum(filterbank, axis=0)
    P_KID = np.nansum(psd_running, axis=0) * dF_sky 
    n_branches = np.unique(index_branches).size - 1

    # Initialise branches
    eta_branches = []
    for n in range(n_branches):
        eta_branches.append(np.ones(F_sky.size))

    branch_counter = 1
    eta_use_flag = False
    for idx_branch, eta_stage in zip(index_branches, eta_cascade):
        if idx_branch == 0:
            break
        
        if idx_branch > branch_counter:
            eta_use = 1 - eta_stage
            eta_use_flag = True
            branch_counter += 1

        for i in range(idx_branch):
            if i == n_branches - 1 and eta_use_flag: # We are at last current branch AND it is start of that branch
                eta_branches[i] *= eta_use
                eta_use_flag = False
            else:
                pass
                eta_branches[i] *= eta_stage

    # Before averaging over filterbank, sum contribution over branches.
    eta_fwd = average_over_filterbank(np.nansum(eta_branches, axis=0), filterbank) 

    # Average eta_inst over filter shapes
    eta_inst = average_over_filterbank(eta_inst, filterbank)
    eta_window = average_over_filterbank(eta_window, filterbank)

    # For eta_sw, also smooth eta_atm over filter response
    eta_atm = average_over_filterbank(eta_atm, filterbank) 

    NEP = np.sqrt(np.nansum(photon_NEP2_kid(F_sky[:,None], psd_running), axis=0) * dF_sky) * KID_excess_noise_factor
    NEP_inst = NEP / eta_inst  # Instrument NEP    

    # ##############################################################
    # 2. Calculating source coupling and sensitivtiy (MDLF and NEFD)
    # ##############################################################

    # Efficiencies
    # .........................................................

    Ag = np.pi * (D_tel / 2.0) ** 2.0  # Geometric area of the telescope
    #omega_mb = np.pi * theta_maj * theta_min / np.log(2) / 4  # Main beam solid angle
    #omega_a = omega_mb / eta_mb  # beam solid angle
    #Ae = (c / (F/350)) ** 2 / omega_a  # Effective Aperture (m^2): lambda^2 / omega_a
    #eta_a = Ae / Ag  # Aperture efficiency

    # Here, I define the illumination efficiency as the product of taper and Ruze efficiencies
    eta_illum = eta_taper * eta_ruze(F_sky, s_rms)
    
    # Calculate aperture efficiency using illumination efficiency and accumulated stages that terminate on-sky
    eta_ap *= eta_illum

    eta_ap = average_over_filterbank(eta_ap, filterbank)
    eta_illum = average_over_filterbank(eta_illum, filterbank)

    # Coupling from the "S"ource to outside of "W"indow
    # Note that not the aperture but illumination efficiency are used, because the forward efficiency includes the stages going to atmosphere
    eta_pol = 0.5  # Instrument is single polarization
    eta_sw = eta_pol * eta_atm * eta_illum * eta_fwd  # Source-Window coupling

    # NESP: Noise Equivalent Source Power (an intermediate quantitiy)
    # .........................................................

    NESP = NEP_inst / eta_sw  # Noise equivalnet source power

    # NEF: Noise Equivalent Flux (an intermediate quantitiy)
    # .........................................................

    # From this point, units change from Hz^-0.5 to t^0.5
    # sqrt(2) is because NEP is defined for 0.5 s integration.

    NEF = NESP / Ag / np.sqrt(2)  # Noise equivalent flux

    # If the observation is involves ON-OFF sky subtraction,
    # Subtraction of two noisy sources results in sqrt(2) increase in noise.

    if on_off:
        NEF *= np.sqrt(2)

    NET = NEP / (np.sqrt(2) * eta_inst * eta_fwd * W_F_cont * k)

    # MDLF (Minimum Detectable Line Flux)
    # .........................................................

    # Note that eta_IBF does not matter for MDLF because it is flux.

    MDLF = NEF * snr / np.sqrt(obs_hours * on_source_fraction * 60.0 * 60.0)

    # NEFD (Noise Equivalent Flux Density)
    # .........................................................

    spectral_NEFD = NEF / W_F_spec
    continuum_NEFD = NEF / W_F_cont  # = spectral_NEFD * eta_IBF < spectral_NEFD

    # Equivalent Trx
    # .........................................................

    Trx = NEP_inst / k / np.sqrt(2 * W_F_cont) - average_over_filterbank(T_from_psd(F_sky, psd_in_front_of_cryo), filterbank)  # assumes RJ!

    # Photon occupation number
    # .........................................................
    n_ph = psd_KID / (h * F)

    # ############################################
    # 3. Output results as dictionary
    # ############################################

    return {
            "F_KID"         : F,
            "F_sky"         : F_sky,
            "EL"            : EL,
            "PWV"           : pwv,
            "R"             : R,
            "filterbank"    : filterbank,
            "W_F_spec"      : W_F_spec,
            "W_F_cont"      : W_F_cont,
            "eta_atm"       : eta_atm,
            "eta_ap"        : eta_ap,
            "eta_fwd"       : eta_fwd,
            "eta_sw"        : eta_sw,
            "eta_window"    : eta_window,
            "eta_inst"      : eta_inst,
            "Tb_sky"        : T_from_psd(F, psd_sky),
            "psd_KID"       : psd_KID,
            "P_KID"         : P_KID,
            "NEP_KID"       : NEP,
            "NEP_inst"      : NEP_inst,
            "NET_sky"       : NET,
            "NEFD_line"     : spectral_NEFD,
            "NEFD_continuum": continuum_NEFD,
            "MDLF"          : MDLF,
            "equivalent_Trx": Trx,
            "n_ph"          : n_ph,
            "W_F_cont"      : W_F_cont
            }



# standard library
from typing import List, Union


# dependent packages
import numpy as np


# type aliases
ArrayLike = Union[np.ndarray, List[float], List[int], float, int]


# constants
h = 6.62607004 * 10**-34  # Planck constant
k = 1.38064852 * 10**-23  # Boltzmann constant
e = 1.60217662 * 10**-19  # electron charge
c = 299792458.0  # velocity of light


# main functions
def rad_trans(rad_in: ArrayLike, 
              medium: ArrayLike, 
              eta: ArrayLike) -> ArrayLike:
    """Calculates radiation transfer through a semi-transparent medium.

    One can also use the same function for Johnson-Nyquist PSD
    (power spectral density) instead of temperature.

    Parameters
    ----------
    rad_in
        Brightness temperature (or PSD) of the input. Units: K (or W/Hz).
    medium
        Brightness temperature (or PSD) of the lossy medium. Units: K (or W/Hz).
    eta
        Transmission of the lossy medium. Units: K (or W/Hz).

    Returns
    -------
    Brightness temperature (or PSD) of the output.

    """

    if len(eta.shape) > 1:
        rad_in = rad_in[:,None]
        medium = medium[:,None]
    return eta * rad_in + (1 - eta) * medium


def T_from_psd(F: ArrayLike, 
               psd: ArrayLike, 
               method: str = "Planck") -> ArrayLike:
    """Calculate Planck temperature from the PSD frequency (frequencies).

    Parameters
    ----------
    F
        Frequency. Units: Hz.
    psd
        Power Spectral Density. Units: W / Hz.
    method
        Default: 'Planck'. Option: 'Rayleigh-Jeans'.

    Returns
    --------
    Planck temperature. Units: K.

    """
    if method == "Planck":
        return h * F / (k * np.log(h * F / psd + 1.0))
    elif method == "Rayleigh-Jeans":
        return psd / k
    else:
        raise ValueError("Method should be Planck or Rayleigh-Jeans.")


def johnson_nyquist_psd(F: ArrayLike, T: ArrayLike) -> ArrayLike:
    """Johnson-Nyquist power spectral density.

    Don't forget to multiply with bandwidth to caculate the total power in W.

    Parameters
    ----------
    F
        Frequency. Units: Hz.
    T
        Temperature. Units: K.

    Returns
    --------
    Power Spectral Density. Units: W / Hz.

    """
    return h * F * nph(F, T)


# helper functions
def nph(F: ArrayLike, T: ArrayLike) -> ArrayLike:
    """Photon occupation number of Bose-Einstein Statistics.

    If it is not single temperature, use nph = Pkid / (W_F * h * F).

    Parameters
    ----------
    F
        Frequency. Units: Hz.
    T
        Temperature. Units: K.

    Returns
    --------
    Photon occupation number. Units: None.

    """
    return 1.0 / np.expm1(h * F / (k * T))



"""Sensitivity calculator for DESHIMA-type spectrometers.

FAQ
-----
Q.  Where is the point-source coupling phase and amplitude loss
    due to the mismatch between the
    beam in radiation and reception, that Shahab calculates
    at the lens surface?

A.  It is included in the main beam efficiency.
    These losses reduce the coupling to a point source,
    but the power (in transmission) couples to the sky.

"""

# flake8: noqa
__version__ = "0.1.4"


# modules
from . import atmosphere
from . import instruments
from . import physics
from . import simulator



# standard library
from typing import List, Union, Tuple


# dependent packages
import numpy as np
from .physics import c, e, h,johnson_nyquist_psd 


# type aliases
ArrayLike = Union[np.ndarray, List[float], List[int], float, int]


# constants
Delta_Al = 188.0 * 10**-6 * e  # gap energy of Al
eta_pb = 0.4  # Pair breaking efficiency
eta_Al_ohmic_850 = 0.9975  # Ohmic loss of an Al surface at 850 GHz.
# Shitov+, ISSTT2008. https://www.nrao.edu/meetings/isstt/papers/2008/2008263266.pdf


# main functions
def D2HPBW(F: ArrayLike) -> ArrayLike:
    """Get half-power beam width of DESHIMA 2.0 at given frequency (frequencies).

    Parameters
    ----------
    F
        Frequency. Units: Hz.

    Returns
    -------
    hpbw
        Half-power beam width. Units: radian.

    """
    return 29.0 * 240.0 / F * np.pi / 180.0 / 60.0 / 60.0


def eta_ruze(F: ArrayLike, 
             sigma: float) -> ArrayLike:
    """Get main-beam efficiency by Ruze's equation.

    Parameters
    ----------
    F
        Frequency. Units: Hz.
    sigma
        Surface error. Units: m.

    Returns
    -------
    eta_ruze
        Ruze efficiency. Units: None.

    """
    return np.exp(-((4.0 * np.pi * sigma * F / c) ** 2.0))


def photon_NEP2_kid(F: ArrayLike, 
                   psd_KID: ArrayLike) -> ArrayLike:
    """NEP squared of the KID, with respect to the absorbed power.
    Note that the square of the NEP is returned, as this quantity is more convenient for the calculation.
    Also, we do not multiply by dF here, as we do that further downstream outside of the integral.

    Parameters
    -----------
    F
        Frequency of the signal responsible for loading. Units: Hz.
    psd_KID
        Power spectral density entering MKID the KID. Units: W / Hz.

    Returns
    -------
    NEP2_kid
        Noise-equivalent power squared of the KID.
    """
    poisson_term = 2 * psd_KID * h * F
    bunching_term = 2 * psd_KID**2
    r_term = 4 * Delta_Al * psd_KID / eta_pb
    return poisson_term + bunching_term + r_term

def window_trans(
    F: ArrayLike,
    thickness: ArrayLike,
    tandelta: float,
    neff: float,
    AR: bool,
    T_parasitic_refl: float,
    T_parasitic_refr: float
) -> Tuple[ArrayLike, ArrayLike]:
    """Calculates the transmission/reflection from dielectric elements, such as lenses and windows.

    Parameters
    ----------
    F
        Frequency. Units: Hz.
    thickness
        Thickness of the lens/window. Units: m.
    tandelta
        Loss tangent of lens/window dielectric.
    neff
        Refractive index of lens/window dielectric. Units : None.
    AR
        Whether the lens/window is supposed to be coated by Anti-Reflective coating (True) or not (False).
    T_parasitic_refl
        Temperature of parasitic source seen in reflection, w.r.t. instrument.
    T_parasitic_refr
        Temperature of parasitic source seen in refraction..

    Returns
    -------
    List containing list of arrays of efficiencies as first element, and list of arrays of psd's seen by each stage.
    """

    eta = []
    psd = []
    
    refl = ((1 - neff) / (1 + neff)) ** 2 * np.ones(F.size)
    psd_refl = johnson_nyquist_psd(F, T_parasitic_refl)
    psd_refr = johnson_nyquist_psd(F, T_parasitic_refr)

    if AR == False:
        eta.append(1 - refl)
        psd.append(psd_refl)
    
    alpha = (2 * np.pi * F / c) * neff * tandelta
    eta_dielectric = np.exp(-alpha * thickness)

    eta.append(eta_dielectric)
    psd.append(psd_refr)

    if AR == False:
        eta.append(1 - refl)
        psd.append(psd_refl)

    return eta, psd

def average_over_filterbank(array_to_average: ArrayLike,
                            filterbank: ArrayLike) -> ArrayLike:
    """Averages a quantity defined on F_sky over as filterbank.

    Parameters
    ----------
    array_to_average
        Array defined over F_sky to average
    filterbank
        2D array containing filterbank

    Returns
    -------
    Averaged array.
    """
    sh_f = filterbank.shape
    assert array_to_average.size == sh_f[0]

    array_tiled = np.tile(array_to_average, (sh_f[-1], 1)) * filterbank.T
    return np.nansum(array_tiled, axis=1) / np.nansum(filterbank, axis=0)



import numpy as np
from typing import List, Union, Dict, Tuple

ArrayLike = Union[np.ndarray, List[float], List[int], float, int]

def Lorentzian(gamma: float, 
               F_sky: ArrayLike, 
               F: ArrayLike) -> ArrayLike:
    """Generate Lorentzians.

    Parameters
    ----------
    gamma
        FWHM of Lorentzian. Units: Hz.
    F_sky
        Array with sky frequencies. Units: Hz.
    F
        Array with center frequencies of Lorentzians.
    
    Returns
    ----------
    2D array with Lorentzians.
    """

    return (gamma**2 / ((F_sky.T - F)**2 + gamma**2))

def generateFilterbankFromR(R: Union[float, int], 
                            F: ArrayLike, 
                            thres: float = 0.0001) -> Tuple[ArrayLike, ArrayLike, float]:
    """Generate a filterbank containing pure Lorentzians.
    
    Parameters
    ----------
    R
        Resolving power of filterbank. Units: none.
    F
        Array with filter frequencies. Units: Hz.
    thres
        Threshold value to use in converge of F_sky
    
    Returns
    ----------
    2D array with filter efficiencies, 1D array with sky frequencies, and stepsize of sky frequencies.. 
    """

    gamma = F / (2 * R)

    mu_low = -gamma[0] * (thres**(-1) - 1)**(1/2) + F[0]
    mu_upp = gamma[-1] * (thres**(-1) - 1)**(1/2) + F[-1]

    area = 0.1
    nF_sky = 750

    area0 = 1e99
    area = 0

    check_a = lambda x, y : np.absolute((x - y) / x) - 1 > thres

    while check_a(area0, area):
        nF_sky *= 2
        F_sky = np.tile(np.linspace(mu_low, mu_upp, nF_sky), (2, 1))
        dF_sky = np.mean(np.diff(F_sky, axis=-1))
        filters = Lorentzian(gamma[0::F.size-1], F_sky, F[0::F.size-1])
        area0 = area
        area = np.sum(filters)

    F_sky = np.tile(np.linspace(mu_low, mu_upp, nF_sky), (F.size, 1))

    filterbank = 4 / np.pi * Lorentzian(gamma, F_sky, F)
    
    #if cutoff := instrumentDict.get("cutoff") is not None:
    filterbank[F_sky[0,:] < 90e9,:] = 0
    
    return filterbank, F_sky[0,:], np.mean(np.diff(F_sky[0,:]))



import numpy as np
import os
import yaml
import csv

from typing import List, Union, Dict, Tuple

from .instruments import eta_Al_ohmic_850, window_trans
from .physics import johnson_nyquist_psd

def eta_Al_ohmic(F_sky: np.ndarray) -> np.ndarray: 
    """Calculate Ohmic losses for aluminium over array of sky frequencies.
    
    Parameters
    ----------
    F_sky
        Numpy array containing sky frequencies. Units: GHz
    
    Returns
    ----------
    Array with eta values for Ohmic losses.
    """

    return 1.0 - (1.0 - eta_Al_ohmic_850) * np.sqrt(F_sky / 850)

def sizer(eta: Union[np.ndarray, float], 
           F_sky: np.ndarray, 
           F_eta: np.ndarray = None
) -> np.ndarray:
    """Resize efficiency term to new size.

    Used to vectorize or interpolate on efficiency terms.
    If efficiency is a scalar, an array is returned with the same size as F_sky.
    If efficiency is an array with different size then F_sky, an array containing frequencies at which eta is evaluated should also be passed.
    A 1D interpolation on F_sky is then performed to evaluate eta on F_sky.
    If efficiency is array with same size as F_sky, it is returned as-is. 
    Responisibility to verify if the efficiencies are evaluated on the same frequencies as present in F_sky is placed on the user.

    Parameters
    ----------
    eta
        Efficiency term.
    F_sky
        Numpy array containing sky frequencies. Units: GHz
    F_eta
        Numpy array containing frequencies at which eta is evaluated.
        Should only be passed when 1D interpolation is required and defaults to None.
    
    Returns
    ----------
    Array with eta values, depending on input (see above).
    """

    if not hasattr(eta, "__len__"):
        return eta * np.ones(F_sky.size)

    elif F_eta is not None:
        idx_sorted = np.argsort(F_eta)
        return np.interp(F_sky, 
                         F_eta[idx_sorted], 
                         eta[idx_sorted])

    else:
        return eta

def read_from_folder(cascade_folder: str,
                     yaml_name: str = "cascade.yaml"
                     ) -> List[Dict[any, any]]:
    """
    Generate a cascade list from a cascade folder.
    The folder should contain a YAML file containing the cascadelist.
    Any vector-valued efficiency terms should be provided inside the folder as a CSV file, with the first column containing frequencies at which the terms are evaluated and the second column containing the terms themselves.
    Then, the CSV can be referenced inside the YAML by passing the CSV name (including .csv) to the `eta_coup` field inside the YAML.

    Parameters
    ----------
    cascade_folder
        String containing path to folder containing cascade YAML and any related CSV files.

    yaml_name
        String containing the name of the YAML file containing the cascade.
        Defaults to 'cascade.yaml'.

    Returns
    ----------
    List containing the cascade.
    """

    assert(os.path.exists(cascade_folder))
    assert(os.path.exists(yaml_path := os.path.join(cascade_folder, 
                                       yaml_name)))

    with open(yaml_path) as stream:
        try:
            cascade_list = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)

    for stage in cascade_list:
        for key, item in stage.items():
            if isinstance(item, str):
                if item.endswith(".csv"):
                    assert(os.path.exists(csv_path := os.path.join(cascade_folder,
                                                                   item)))
                    freq = []
                    vals = []
                    with open(csv_path, 'r', newline='') as csvfile:                
                        reader = csv.reader(csvfile, delimiter=',')
                        for row in reader:
                            freq.append(float(row[0]))
                            vals.append(float(row[1]))

                    stage[key] = (np.array(freq), np.array(vals))

    return cascade_list

def save_cascade(cascade_list: List[Dict[any, any]],
                 save_folder: str,
                 yaml_name: str = "cascade") -> None:
    if not os.path.exists(save_folder):
        os.makedirs(save_folder)

    name_index = 0

    cascade_to_write = []

    for stage in cascade_list:
        stage_dict = {}

        for key, item in stage.items():
            if isinstance(item, tuple):
                assert(item[0].size == item[1].size)
                np.savetxt(os.path.join(save_folder, 
                                        f"{name_index}.csv"), 
                           np.column_stack(item),
                           delimiter = ",")

                stage_dict[key] = f"{name_index}.csv"
                
                name_index += 1

            else:
                stage_dict[key] = item

        cascade_to_write.append(stage_dict)

    with open(os.path.join(save_folder, f"{yaml_name}.yaml"), 'w') as outfile:
              yaml.dump(cascade_to_write, outfile)

def get_cascade(cascade_list: List[Dict[any, any]],
                F_sky: np.ndarray
                ) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate a cascade list, consisting of efficiency and psd per stage.

    Parameters
    ----------
    cascade_list
        List containing, per element, the efficiency and coupling temperature of each stage in the cascade.
        For reflective stages, the dictionary should contain either:
            - A single eta and temperature
            - A tuple with efficiencies and frequencies at which these are defined, and a temperature
        
        For refractive stages, the dictionary should contain:
            - thickness of dielectric in meters, loss tangent, effective refractive index, whether to use AR coating, temperature seen in reflection coming from the ISS, and temperature seen in refraction.

    F_sky
        Array with sky frequencies. Units: GHz.
    
    Returns
    ----------
    List with list of arrays containing efficiencies as first element, and list containing arrays of psd as second element. 
    """


    group_list = []
    
    for cascade in cascade_list:
        group_list.append(cascade.get("group"))
    
    group_list_red_uniq = list(dict.fromkeys([x for x in group_list if x is not None]))

    idx_group_list = [i if label == None else label for i, label in enumerate(group_list)]
    cascade_type_list = np.array([0 if x.get("eta_coup") is not None else 1 for x in cascade_list])
    idx_group_list_uniq = list(dict.fromkeys(idx_group_list))

    to_delete = []
    for group_label in group_list_red_uniq:
        index_list = [i for i, label in enumerate(idx_group_list) if label == group_label]
        idx_group_list_uniq[idx_group_list_uniq.index(group_label)] = index_list

        to_delete.extend(index_list[:-1])

    cascade_type_list_uniq = np.delete(cascade_type_list, to_delete)
    idx_group_list_uniq = [[x] if hasattr(x, "__len__") == False else x for x in idx_group_list_uniq] 

    # Now get eta for groups

    group_running = None
    all_eta = []
    all_psd = []
    use_for_eta_inst = []
    use_for_eta_ap = []
    eta_inst_flag = 0

    for casc_t, idx_group in zip(cascade_type_list_uniq, idx_group_list_uniq):
        # This cascade group is reflective
        eta_ap_flag = 0
        if casc_t == 0:
            eta_grouped = np.ones(F_sky.size)

            if (T_casc := cascade_list[idx_group[0]].get("T_parasitic")) == "atmosphere":
                all_psd.append(T_casc) # Group couples to atmosphere: calculate psd in rad trans loop.
                eta_ap_flag = 1
            else:
                all_psd.append(johnson_nyquist_psd(F_sky, T_casc)) # Calculate psd for T_parasitic

            for idx_g in idx_group:
                eta_interp_flag = False
                
                if isinstance(eta := cascade_list[idx_g].get("eta_coup"), tuple): # eta_coup is a tuple; array of eta and frequencies
                    assert(len(eta) == 2)
                    
                    eta, F_eta = eta

                    assert(eta.size == F_eta.size)
                    
                    eta_interp_flag = True

                elif eta == "Ohmic-Al": # generate vector with eta of Aluminium
                    eta = eta_Al_ohmic(F_sky) 

                elif eta_interp_flag:
                    eta_grouped *= sizer(eta, F_sky, F_eta)
                else:
                    eta_grouped *= sizer(eta, F_sky)
            
            all_eta.append(eta_grouped)
            
            use_for_eta_inst.append(eta_inst_flag)
            use_for_eta_ap.append(eta_ap_flag)
        
        if casc_t == 1:
            for idx_g in idx_group:
                if (casc := cascade_list[idx_g]).get("cryo_window_flag"):
                    eta_inst_flag = 1
                etas, psds = window_trans(F_sky,
                                          casc.get("thickness"), 
                                          casc.get("tandelta"), 
                                          casc.get("neff"), 
                                          casc.get("AR"), 
                                          casc.get("T_parasitic_refl"), 
                                          casc.get("T_parasitic_refr")) 

                all_eta.extend(etas)
                all_psd.extend(psds)
        
                use_for_eta_inst.extend([eta_inst_flag for _ in range(len(etas))])
                use_for_eta_ap.extend([eta_ap_flag for _ in range(len(etas))])
    
    return all_eta, all_psd, use_for_eta_inst, use_for_eta_ap



import numpy as np
import os

Tp_amb = 273 # Ambient temperature
Tp_cabin = 290
Tp_co = 4
Tp_chip = 0.12

eta_M1_spill = 0.99
eta_M2_spill = 0.9
eta_WO = 0.99
eta_co = 0.65
eta_lens_antenna_rad = 0.81
eta_circuit = 0.32

d = 8e-3
tandelta = 4.805e-4
neffHDPE = 1.52
window_AR = True
#window_AR = False

F0 = 220 * 1e9
nF = 350
R = 500
F = F0 * (1 + 1/R)**np.arange(nF)

F_QO, eta_QO = np.load(os.path.join("tifuun_sensitivity", "data", "QOfilterstack.npy"))

cascade_list = [
        {"name" : "spillover M1",
             "eta_coup" : eta_M1_spill,
             "T_parasitic" : Tp_amb,
             "group" : "amb"}, # Spillover cascade on M1
            {"name" : "Ohmic M1", 
             "eta_coup" : "Ohmic-Al", 
             "T_parasitic" : Tp_amb,
             "group" : "amb"}, # Ohmic losses on M1
            {"name" : "Ohmic M2",
             "eta_coup" : "Ohmic-Al",
             "T_parasitic" : Tp_amb,
             "group" : "amb"}, # Ohmic losses on M2
            {"eta_coup" : eta_M2_spill, 
             "T_parasitic" : "atmosphere"}, 
            {"eta_coup" : eta_WO,
             "T_parasitic" : Tp_cabin,
             "group" : "cabin"}, # Spillover on WO
            {"eta_coup" : "Ohmic-Al",
             "T_parasitic" : Tp_cabin,
             "group" : "cabin"}, # Ohmic losses WO1
            {"eta_coup" : "Ohmic-Al",
             "T_parasitic" : Tp_cabin,
             "group" : "cabin"}, # Ohmic losses on WO 2
            {"thickness" : d, 
             "tandelta" : tandelta, 
             "neff" : neffHDPE,
             "AR" : window_AR,
             "T_parasitic_refl" : Tp_co,
             "T_parasitic_refr" : Tp_cabin,
             "cryo_window_flag" : True}, # Cryostat window
            {"eta_coup" : (F_QO, eta_QO),
             "T_parasitic" : Tp_co,
             "group" : "cryo"},
            {"eta_coup" : "Ohmic-Al",
             "T_parasitic" : Tp_co,
             "group" : "cryo"}, # Ohmic losses on CO 1
            {"eta_coup" : eta_co,
             "T_parasitic" : Tp_co,
             "group" : "cryo"}, 
            {"eta_coup" : "Ohmic-Al",                        
             "T_parasitic" : Tp_co,                          
             "group" : "cryo"}, # Ohmic losses on CO 2       
            {"eta_coup" : eta_lens_antenna_rad * eta_circuit,
             "T_parasitic" : Tp_chip}
            ]

instrument = {                         
        "f_ch"      : F,               
        "R"         : R,               
        "eta_IBF"   : 0.5,             
        "KID_excess_noise_factor" : 1.1
        }                              

telescope = {               
        "D_tel"     : 10,   
        "s_rms"     : 42e-6,
        "eta_taper" : 0.7,  
        "EL0"       : 60    
        }                   




# standard library
from pathlib import Path
from typing import Callable, List, Union


# dependent packages
import numpy as np
import pandas as pd
from scipy.interpolate import RectBivariateSpline

# type aliases
ArrayLike = Union[np.ndarray, List[float], List[int], float, int]

# main functions
def eta_atm_func(
    F: ArrayLike, 
    pwv: float, 
    EL: float) -> ArrayLike:
    """Calculate eta_atm as a function of F by interpolation.

    Note that currently, the function can only accept scalar values for the precitpitable water vapor (pwv).

    Parameters
    ----------
    F
        Frequency of the astronomical signal.
        Units: Hz.
    pwv
        Precipitable water vapour. Units: mm.
    EL
        Telescope elevation angle. Units: degrees.

    Returns
    -------
    Atmospheric tranmsmission, interpolated on F and pwv. Units: None.
    """

    eta_atm_df = pd.read_csv(
        Path(__file__).parent / "data" / "atm.csv",
        skiprows=4,
        sep='\s+',
        header=0,
    )
    
    x = eta_atm_df["F"].values * 1e9
    y = np.array(list(eta_atm_df)[1:]).astype(np.float64)
    z = eta_atm_df.iloc[:, 1:].values

    eta_atm_interp = RectBivariateSpline(x, y, z, kx=1, ky=1)(F, pwv, grid=True)
    eta_el = eta_atm_interp ** (1 / np.sin(EL * np.pi / 180))
    return eta_el

